import time
from typing import Callable

from pyselenscrapr.ScrapingLogic import ScrapingLogic
from pyselenscrapr.ScrapingStep import ScrapingStep

class ScrapingLogicIterator(ScrapingLogic):
    def __init__(self, logic: ScrapingLogic, element, index):
        super().__init__(logic._driver, logic._bot)
        self._element = element
        self._index = index

    def index(self):
        return self._index

    def element(self):
        return self._element

class ScrapingStepIteration(ScrapingStep):
    pass



class ScrapingStepLoop(ScrapingStep):
    def __init__(self, name: str,
                 iteration_callback: Callable[[ScrapingLogic], any],
                 iteration_steps: list[ScrapingStep],
                 iterations: int =None,
                 retry_count: int = 3):
        super().__init__(name, lambda x: self.execute(x))
        self._iteration_callback = iteration_callback
        self._iteration_steps = iteration_steps
        self.current_iteration = 0

    def execute(self, logic: ScrapingLogic):
        self.elements = self._iteration_callback(logic)

        index = 0
        for element in self.elements:
            try:
                for step in self._iteration_steps:
                    l = ScrapingLogicIterator(logic, element, index)
                    step.execute(l)
                    time.sleep(2)
            except Exception as e:
                logic._on_exception(e, step)
            index += 1


import random
import time
from typing import Callable

from pyselenscrapr.ScrapingLogic import ScrapingLogic
from pyselenscrapr.ScrapingStep import ScrapingStep, IScrapingStep


class ScrapingStepPaginationMode:
    RandomPages = 1
    AllPages = 2

class ScrapingStepPagination(ScrapingStep):
    def __init__(self, name: str,
                 execute: Callable[[IScrapingStep], any],
                 goto_page: Callable[[IScrapingStep, int], None],
                validate_page: Callable[[IScrapingStep, int], bool],
                 pagination_mode: ScrapingStepPaginationMode,
                 page_count: Callable[[IScrapingStep], int],
                 exit_bot_when_errored: bool = False,):
        super().__init__(name, execute, exit_bot_when_errored=exit_bot_when_errored)
        self._executionList = []
        self._goto_page = goto_page
        self._pagination_mode = pagination_mode
        self._validate_page = validate_page
        self._page_count = page_count
        self._min_wait_time = 2
        self._max_wait_time = 10

    def _fill_all_pages(self):
        list = []
        # generate a list of random page numbers until _page_count_value
        for i in range(0, self._page_count_value+1):
            list.append(None)
        return list

    def finished(self):
        for i in self._executionList:
            if i is None:
                return False
        return True

    def _get_next_page(self):
        # TODO: remove because of testing
        return 500
        """
        for i in range(1, len(self._executionList)):
            if self._pagination_mode == ScrapingStepPaginationMode.RandomPages:
                r = random.randint(1, len(self._executionList)-1)
                if self._executionList[r] is None:
                    return r
            else:
                if self._executionList[i] is None:
                    return i
                """
        return None

    def _sleep(self, t):
        time.sleep(t)

    def can_retry(self):
        return False

    def retry(self):
        self._sleep(1)

    def sleep_random(self):
        t = random.randint(self._min_wait_time, self._max_wait_time)
        self._sleep(t)

    def execute(self, logic):
        self._page_count_value = self._page_count(logic)
        self._executionList = self._fill_all_pages()

        while not self.finished():
            next_page = self._get_next_page()
            self.log("Scraping page " + str(next_page) + " of " + str(self._page_count_value) + " pages")
            self.log("Pages left: " + str(len([x for x in self._executionList if x is None])))

            if next_page is None:
                break

            retry = 3
            success = False
            l = ScrapingLogic(logic._driver, logic._bot)
            for i in range(0, retry):
                try:
                    self._goto_page(l, next_page)
                    l.sleep(1)
                    if self._validate_page(l, next_page):
                        success = True
                        break
                except Exception as e:
                    self.log("Error: " + str(e))
                    l.sleep(1)

            if not success:
                self.raise_exception("Could not navigate to page " + str(next_page)+" after "+str(retry)+" attempts.")

            try:
                l = ScrapingLogic(logic._driver, logic._bot)
                self._execute(l)
                l.take_screenshot(self)
                l.sleep(1)
            except Exception as e:
                self.log("Error on try to scrape logic" + str(e))

            self._executionList[next_page] = True
            self.sleep_random()








import time
from selenium import webdriver
from selenium.webdriver import Keys
from selenium.webdriver.common.by import By

from pyselenscrapr.ScrapingBot import ScrapingBot, TakeScreenshotModes
import logging as log

from pyselenscrapr.ScrapingLogic import ScrapingLogic
from pyselenscrapr.ScrapingStep import ScrapingStep, ScrapingStepInterval, ScrapingStepErrorHandling

ff_options = webdriver.FirefoxOptions()
ff_options.add_argument('--no-sandbox')
ff_options.add_argument("--lang=en-GB")
ff_options.add_argument('--disable-dev-shm-usage')

"""
# set proxy to firefox
ff_options.set_preference('network.proxy.type', 1)
ff_options.set_preference('network.proxy.socks', "localhost")
ff_options.set_preference('network.proxy.socks_port', 9050)
"""

driver = webdriver.Remote("http://localhost:4444/wd/hub", options=ff_options)

log.basicConfig(level=log.DEBUG)
robot = ScrapingBot(driver, take_screenshots_mode=TakeScreenshotModes.Always)

try:

    search_command = "wikipedia matrix movies"

    robot.add_step(ScrapingStep("Check if there is a visible 'Accept all' or 'decline all' button",
                                lambda l: l.get_best_element("//button//div[contains(text(), 'All')]").click(),
                                interval=ScrapingStepInterval.BeforeAnyStep,
                                error_handling=ScrapingStepErrorHandling.Ignore,
                                can_execute=lambda l: l.current_url and "google" in l.current_url and l.is_visible("//button//div[contains(text(), 'All')]"),
                                was_executed=lambda l: not l.is_visible("//button//div[contains(text(), 'All')]"),),
                   step_group="default")

    step1 = robot.add_step(ScrapingStep("Go to google.de and check if textarea is ready",
                                        lambda l: l.get("https://www.google.de"),
                                        was_executed=lambda l: l.element_exists("//textarea")),
                           step_group="default")

    step2 = step1.next_step(ScrapingStep("Send the search command to the textarea",
                                         lambda l: l.send_keys_to_element("//textarea", search_command),
                                            was_executed=lambda l: l.element_text("//textarea") == search_command)
)

    step3 = step2.next_step(ScrapingStep("Send the ENTER key to the textarea",
                                            lambda l: l.send_keys_to_element("//textarea", Keys.ENTER),
                                            was_executed=lambda l: l.element_count("//h3") > 3))


    step4 = step3.next_step(ScrapingStep("Click on the first search result",
                                            lambda l: l.get_best_element("//h3").click(),
                                            was_executed=lambda l: "google" not in l.current_url))

    step5 = step4.next_step(ScrapingStep("Extrac tall tables from the page",
                                            lambda l: l.set_data("tables", l.convert_tables_to_df(l.get_all_elements("//table"))),
                                            was_executed=lambda l: l.has_data("tables"))
                            )

    robot.set_warning_handler(lambda w: log.warning("Warning: ", w))
    robot.set_exception_handler(lambda e: log.error("Exception: ", e))

    finished_success = robot.run()

    if finished_success:
        print(robot.get_data("tables"))
    else:
        print("The robot did not finish successfully")

except Exception as e:
    log.error(e)
finally:
    driver.quit()


import json
import logging
from abc import ABC
import requests

# interface IScrapingBackend with the method "saveData" and "errorHandling" and "notify"
class IScrapingBackend(ABC):
    """
    This is a interface that represents a backend for the scraping process.
    """
    def saveData(self, data: dict, key: str = None):
        raise NotImplementedError

    def errorHandling(self, error: Exception, debugData=None):
        raise NotImplementedError

    def notify(self, message: str):
        pass


class ScrapingBackendWebhook(IScrapingBackend):
    """
    This class is used to send the data you scraped to a webhook.
    """
    _url = None
    def __init__(self, url, error_route="/error", notify_route="/notify", data_route="/data"):
        self._url = url
        self._error_route = url + error_route
        self._notify_route = url + notify_route
        self._data_route = url + data_route

    def saveData(self, data: dict, key: str = None):
        try:
            d = json.dumps(data)
            ret = requests.post(self._data_route, data=d, headers={"Content-Type": "application/json", "Accept": "application/json"})
            logging.debug("data return")
            logging.debug(ret.status_code)
            logging.debug(ret)
        except Exception as e:
            print(e)
            raise e

    def errorHandling(self, error: Exception, debugData=None):
        try:
            d = {"error": str(error)}
            if debugData is not None:
                d["debug"] = debugData
            d_json = json.dumps(d)
            ret = requests.post(self._error_route, data=d_json, headers={"Content-Type": "application/json", "Accept": "application/json"})
            logging.debug("data return")
            logging.debug(ret.status_code)
            logging.debug(ret)
        except Exception as e:
            logging.error(e)
            raise e

    def notify(self, message: str):
        try:
            d = json.dumps({"message": str(message)})
            ret = requests.post(self._notify_route, data=d, headers={"Content-Type": "application/json", "Accept": "application/json"})
            logging.debug("data return")
            logging.debug(ret.status_code)
            logging.debug(ret)
        except Exception as e:
            logging.error(e)
            raise e


from pyselenscrapr.ScrapingStep import IScrapingStep


class ScrapingStepGroup:
    _steps : [IScrapingStep] = []
    name = None

    def __init__(self, name: str, steps: [IScrapingStep]=[]):
        self._steps = steps
        self.name = name

    def add_step(self, step):
        self._steps.append(step)
        step.group = self


import logging as log
import time
from typing import Callable

import numpy as np
import pandas as pd
from typing import Union
from pyselenscrapr.ScrapingBackend import IScrapingBackend
from pyselenscrapr.ScrapingLogic import ScrapingLogic
from pyselenscrapr.ScrapingStep import ScrapingStep, ScrapingStepInterval, ScrapingStepErrorHandling
from pyselenscrapr.ScrapingStepGroup import ScrapingStepGroup

class TakeScreenshotModes:
    """
    This enum is used to define when the bot should take a screenshot of the current page.
    """
    OnError = 2 # take a screenshot when an error occurs
    Always = 1 # take a screenshot for each step
    Never = 0 # never take a screenshot


class ScrapingBot:
    """

    The ScrapingBot class is the main class to be used for creating a scraping bot. It is used to define the steps and
    groups of steps that the bot should execute. The bot can be run by calling the run() method.

    """
    _data = {}
    _driver = None
    _warning_handler = None
    _current_group = None
    _exception_handler = None
    _take_screenshot_on_error = None
    _retry_count = 0
    _repeat_count_till_error = 5
    _repeat_count = 0
    _stepGroups = []
    _max_retries = 3
    _take_screenshots_mode = TakeScreenshotModes.Never
    _screenshot_path = "."
    _backend : IScrapingBackend = None

    def __init__(self, driver, max_retries=3,
                 take_screenshots_mode=TakeScreenshotModes.Never,
                 backend : IScrapingBackend = None,
                 repeat_count_till_error=5):
        self._repeat_count_till_error = repeat_count_till_error
        self._driver = driver
        self._backend = backend
        self._max_retries = max_retries
        self._task_logs = []
        self._take_screenshots_mode = take_screenshots_mode

    def _on_warning(self, w):
        if self._warning_handler is not None:
            self._warning_handler(w)
        else:
            log.warning(w)

    def group_name(self):
        return self._current_group.name

    def backend_notify(self, message):
        if self._backend is not None:
            self._backend.notify(message)

    def _take_screenshot(self, step):
        try:
            self._driver.save_screenshot(self._screenshot_path+self.group_name()+"_"+step.name()+".png")
        except:
            pass

    def _raise_exception(self, message):
        self._on_exception(message, None)
        raise Exception(message)

    def _on_exception(self, e, step):
        self._task_logs.append({
            "time": time.strftime("%Y-%m-%d %H:%M:%S"),
            "message": str(e),
            "step": step.name(),
            "level": "error"
        })
        if self._take_screenshots_mode == TakeScreenshotModes.OnError or \
                self._take_screenshots_mode == TakeScreenshotModes.Always:
            self._take_screenshot(step)
        if self._exception_handler is not None:
            self._exception_handler(e)
        else:
            log.error(e)
        if self._backend is not None:
            self._backend.errorHandling(e, debugData=self._task_logs)

    def set_warning_handler(self, param):
        self._warning_handler = param

    def set_exception_handler(self, param):
        self._exception_handler = param

    def take_screenshot_on_error(self, path):
        self._take_screenshot_on_error = path

    def add_step_group(self, group_name: str):
        group = ScrapingStepGroup(group_name)
        self._stepGroups.append(group)
        return group


    def add_step(self, step_or_callback: ScrapingStep,
                 step_group: ScrapingStepGroup = None,
                 before_validation: Callable[[ScrapingLogic], None]  = None,
                 after_validation: Callable[[ScrapingLogic], None]  = None) -> ScrapingStep:
        group = None
        if isinstance(step_or_callback, ScrapingStep):
            step_or_callback.before_validation = before_validation
            step_or_callback.after_validation = after_validation
        step_or_callback.robot = self

        if step_group is None:
            step_group = "default"
        if step_group is not None and isinstance(step_group, str):
            if len(self._stepGroups) > 0:
                group = next((x for x in self._stepGroups if x.name == step_group), None)
        elif step_group is not None and isinstance(step_group, ScrapingStepGroup):
            group = step_group


        if group is None:
            group = ScrapingStepGroup("default")
            self._stepGroups.append(group)

        group.add_step(step_or_callback)

        return step_or_callback

    def _run_step(self, step, retryInterval=0):
        try:
            if step.can_execute is not None and not step.can_execute(ScrapingLogic(self._driver, self)):
                return
        except Exception as e:
            if step.error_handling() != ScrapingStepErrorHandling.Ignore:
                self._on_exception(e, step)

        try:
            if step.previous_step() is not None and not step.previous_step().was_executed():
                return
        except Exception as e:
            if step.error_handling() != ScrapingStepErrorHandling.Ignore:
                self._on_exception(e, step)

        if step.before_validation is not None:
            try:
                step.before_validation()
            except Exception as e:
                if step.error_handling() != ScrapingStepErrorHandling.Ignore:
                    self._on_exception(e, step)
                else:
                    self._on_warning(e)

        for i in range(self._max_retries):
            try:
                l = ScrapingLogic(self._driver, self)
                step.execute(l)
                if step.is_executed(l):
                    step.set_executed()
                break
            except Exception as e:
                if step.error_handling() != ScrapingStepErrorHandling.Ignore:
                    self._on_exception(e, step)
                else:
                    self._on_warning(e)

                if step.can_retry():
                    step.retry()
                else:
                    if step.exit_bot_when_errored():
                        self._raise_exception("The bot will exit because of an error in the step: "+step.name())

                    break



        if step.after_validation is not None:
            step.after_validation()

    def sleep(self, seconds):
        time.sleep(seconds)

    def _on_debug(self, msg, *args):
        msg = msg+" ".join([str(x) for x in args])
        log.debug(msg)
        self._task_logs.append({
            "time": time.strftime("%Y-%m-%d %H:%M:%S"),
            "message": msg,
            "level": "debug"
        })

    def all_groups_executed(self):
        for group in self._stepGroups:
            for step in group._steps:
                if not step.was_executed():
                    return False
        return True

    def get_next_step(self, step):
        for s in self._current_group._steps:
            if s.interval() != ScrapingStepInterval.Order:
                continue
            if s.was_executed():
                continue
            if s.previous_step() is None:
                return s
            if s.previous_step().was_executed():
                return s


        return None

    def set_current_group(self, group):
        self._current_group = group

    def finished(self):
        return self.all_groups_executed()

    def get_all_steps_by_interval(self, interval):
        steps = []
        for step in self._current_group._steps:
            if step.interval() == interval and not step.was_executed():
                steps.append(step)
        return steps

    def _run_before_step(self, step):
        steps = self.get_all_steps_by_interval(ScrapingStepInterval.BeforeAnyStep)
        for s in steps:
            self._run_step(s)

    def get_next_group(self):
        for i, group in enumerate(self._stepGroups):
            if group == self._current_group:
                if i+1 < len(self._stepGroups):
                    return self._stepGroups[i+1]
        return None

    def _run_after_step(self, step):
        if self._take_screenshots_mode == TakeScreenshotModes.Always:
            self._take_screenshot(step)
        steps = self.get_all_steps_by_interval(ScrapingStepInterval.AfterAnyStep)
        for s in steps:
            self._run_step(s)

    def _run_after_group(self, group):
        pass

    def _is_group_finished(self, group):
        for step in group._steps:
            if not step.was_executed():
                return False
        return True

    def run(self, first_group: Union[str , ScrapingStepGroup]  = None):
        """
        Run the bot and execute all steps in the defined groups.

        :param first_group: This is the name of the first group to start. If it is None we use "default" as the first group.
        :return: True if the bot finished successfully, False otherwise.
        """
        if first_group is None:
            first_group = "default"
        if first_group is not None and isinstance(first_group, str):
            if len(self._stepGroups) > 0:
                group = next((x for x in self._stepGroups if x.name == first_group), None)
        elif first_group is not None and isinstance(first_group, ScrapingStepGroup):
            group = first_group

        if group is None:
            raise Exception("Group not found - please enter a valid group name or object to start the execution from.")
        self.set_current_group(group)
        next_step : ScrapingStep = None
        while not self.finished():
            last_step = next_step

            if self._repeat_count > self._repeat_count_till_error:
                self._on_exception("The same step was repeated too many times", last_step)
                return False

            while not self._is_group_finished(self._current_group):
                next_step = self.get_next_step(next_step)

                if next_step is None:
                    break

                if next_step == last_step:
                    self._repeat_count += 1
                else:
                    self._repeat_count = 0

                self._run_before_step(next_step)

                success = False
                try:
                    self._on_debug("Running step: ", next_step.name())
                    self._run_step(next_step)
                    success = True
                except Exception as e:
                    self._on_debug("Exception: ", e)
                    if next_step.error_handling() != ScrapingStepErrorHandling.Ignore:
                        self._on_exception(e, next_step)

                if success:
                    self._run_after_step(next_step)

                    self.sleep(2)
                    self._on_debug("Finished step: ", next_step.name())
                else:
                    if next_step.exit_bot_when_errored():
                        self._on_exception("The bot will exit because of an error in the step: "+next_step.name(), next_step)
                        return False
                    self._on_debug("Failed step: ", next_step.name()+ " retrying "+str(self._max_retries)+" times.")
                    self.sleep(3)


            self._run_after_group(self._current_group)

            self.sleep(1)
            self.set_current_group(self.get_next_group())

        self.sleep(1)
        return True

    ##############################################################################################################
    # Data handling
    ##############################################################################################################

    def get_converted_data(self, data):
        retdata = {}
        for key, value in data.items():
            if isinstance(value, pd.DataFrame):
                retdata[key] = value.replace({np.nan:None}).to_dict(orient='records')
            else:
                retdata[key] = value

        return retdata

    def save_backend_data(self, data):
        try:
            if self._backend is not None:
                d_converted = self.get_converted_data(data)
                self._backend.saveData(d_converted)
        except Exception as e:
            self._on_warning(e)

    def send_error_to_backend(self, error):
        if self._backend is not None:
            self._backend.errorHandling(error, debugData=self._task_logs)

    def send_data_to_backend(self, key=None, data=None):
        if data is None:
            data = self._data
        if key is not None:
            data = data[key]

        self.save_backend_data(data)

    def set_data(self, key, value, send_to_backend=False):
        self._data[key] = value
        if send_to_backend:
            self.save_backend_data({key: value})

    def append_data(self, key, value, send_to_backend=False):
        if key in self._data:
            # pandas dataframe
            if isinstance(self._data[key], pd.DataFrame):
                # append elements to the dataframe
                self._data[key] = self._data[key]._append(value, ignore_index=True)
            elif isinstance(self._data[key], list):
                self._data[key].append(value)
            elif isinstance(self._data[key], str):
                self._data[key] = [self._data[key], value]
            elif isinstance(self._data[key], int):
                self._data[key] = [self._data[key], value]
            elif isinstance(self._data[key], float):
                self._data[key] = [self._data[key], value]
            elif isinstance(self._data[key], dict):
                self._data[key] = [self._data[key], value]
            elif isinstance(self._data[key], tuple):
                self._data[key] = [self._data[key], value]
            elif isinstance(self._data[key], set):
                self._data[key] = [self._data[key], value]
            # pandas dataframe
            elif isinstance(self._data[key], pd.DataFrame):
                # append elements to the dataframe
                self._data[key] = self._data[key].append(value)
        else:
            self._data[key] = value

        if send_to_backend:
            self.save_backend_data({key: value})

    def has_data(self, key):
        if key in self._data:
            return True
        return False

    def get_data(self, key):
        if key in self._data:
            return self._data[key]
        return None

    def get_task_log(self):
        return self._task_logs



# Configuration file for the Sphinx documentation builder.

# -- Project information
import os
import sys
sys.path.insert(0, os.path.abspath(os.path.join("..", "..")))

project = 'PySelenScrapr'
copyright = '2024, Thoren Lederer'
author = 'donnercody'

release = '0.0.8'
version = '0.0.8'

# -- General configuration

extensions = [
    'sphinx.ext.duration',
    'sphinx.ext.doctest',
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.intersphinx',
    'sphinx.ext.napoleon',
    'sphinx.ext.autosummary',
]

intersphinx_mapping = {
    'python': ('https://docs.python.org/3/', None),
    'sphinx': ('https://www.sphinx-doc.org/en/master/', None),
}
intersphinx_disabled_domains = ['std']

templates_path = ['_templates']

# -- Options for HTML output

html_theme = 'sphinx_rtd_theme'

# -- Options for EPUB output
epub_show_urls = 'footnote'

# autodoc
autodoc_default_options = {
    'members': True,
    'member-order': 'bysource',
    'special-members': True,
    'undoc-members': True,
    'exclude-members': '__weakref__',
    'show-inheritance': True,
}



from typing import Callable

from pyselenscrapr import ScrapingLogic

class ScrapingStepErrorHandling:
    ThrowException = 0
    RetryAndThrowException = 1
    Ignore = 2

#ScrapintStepInterval is an enum
class ScrapingStepInterval:
    Order = 0
    BeforeAnyStep = 1
    AfterAnyStep = 2
    BeforeValidation = 3
    AfterValidation = 4
    BeforeRetry = 5
    AfterRetry = 6
    BeforePagination = 7
    AfterPagination = 8

class ScrapingStepRepeat:
    Repeat = 0
    NoRepeat = 1

# interface IScrapingStep
class IScrapingStep:
    def __init__(self):
        pass

class ScrapingStep(IScrapingStep):
    """
    ScrapingStep is a class that represents a single step in a scraping process.
    """
    childGroups = []
    _hasExecuted = False
    _previous_step = None
    _interval = ScrapingStepInterval.Order
    robot = None
    _repeat = ScrapingStepRepeat.NoRepeat
    _error_handling = ScrapingStepErrorHandling.RetryAndThrowException

    def __init__(self, name: str,
                 execute: Callable[[ScrapingLogic], None],
                 can_execute: Callable[[ScrapingLogic], bool] = None,
                 was_executed: Callable[[ScrapingLogic], bool] = None,
                 before_validation: Callable[[ScrapingLogic], None] = None,
                 retry: Callable[[ScrapingLogic], None] = None,
                 previous_step: IScrapingStep = None,
                 error_handling: ScrapingStepErrorHandling = ScrapingStepErrorHandling.RetryAndThrowException,
                 interval: ScrapingStepInterval = ScrapingStepInterval.Order,
                 repeat: ScrapingStepRepeat = ScrapingStepRepeat.NoRepeat,
                 exit_bot_when_errored: bool = False,
                 retry_count: int = 3):
        """
        Constructor for ScrapingStep

        :param name: a string representing the name of the step - it is used later for debugging purposes
        :param execute: a function that will be executed when the step is executed
        :param can_execute: a function that will be executed to check if the step can be executed
        :param was_executed: a function that will be executed to check if the step was executed
        :param before_validation: a function that will be executed before the step is validated
        :param retry: a function that will be executed if the step fails
        :param previous_step: the previous step in the scraping process
        :param error_handling: an enum representing the error handling strategy
        :param interval: an enum representing the interval at which the step will be executed
        :param repeat: an enum representing if the step should be repeated
        :param retry_count: an integer representing the number of retries
        """
        self._name = name
        self._error_handling = error_handling
        self._repeat = repeat
        self.can_execute = can_execute
        self._previous_step = previous_step
        self._interval = interval
        self._execute = execute
        self._was_executed = was_executed
        self._before_validation = before_validation
        self._retry = retry
        self._exit_bot_when_errored = exit_bot_when_errored
        self._retry_count = retry_count

    def __str__(self):
        return self._name + " " + str(self._interval) + " " + str(self._repeat)


    def name(self) -> str:
        return self._name

    def interval(self) -> ScrapingStepInterval:
        return self._interval

    def next_step(self, step):
        if self.robot is not None:
            step.set_previous_step(self)
            self.robot.add_step(step)
        else:
            raise Exception("The robot is not set")
        return step

    def raise_exception(self, message):
        self.log("Raise exception: " + message)
        raise Exception(message)

    def set_previous_step(self, step):
        self._previous_step = step

    def add_child_group(self, group):
        self.childGroups.append(group)

    def previous_step(self) -> IScrapingStep:
        return self._previous_step

    def can_execute(self) -> bool:
        pass

    def was_executed(self) -> bool:
        return self._hasExecuted

    def execute(self, logic):
        retData = self._execute(logic)
        return retData

    def set_executed(self):
        self._hasExecuted = True

    def before_validation(self):
        pass

    def reset(self):
        self._hasExecuted = False

    def log(self, message):
        if self.robot is not None:
            self.robot._on_debug(message)

    def retry(self):
        if self._retry is not None:
            self._retry()

    def exit_bot_when_errored(self):
        return self._exit_bot_when_errored

    def can_retry(self):
        return True

    def is_executed(self, logic):
        if self._was_executed is not None:
            return self._was_executed(logic)
        return True

    def error_handling(self) -> ScrapingStepErrorHandling:
        return self._error_handling

class ScrapingStepConditional(ScrapingStep):
    pass




import time

import pandas as pd
from selenium.webdriver import Keys
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.wait import WebDriverWait


def tocontainer(func, bot):
    def wrapper(*args, **kwargs):
        result = func(*args, **kwargs)
        return ScrapingLogic(result, bot)
    return wrapper

class ScrapingLogic(object):
    """
    ScraipingLogic is a class that is the main interface between the bot and the selenium driver.
    It contains a lot of helper functions that make it easier to interact with the driver.

    You can use all driver functions and also the functions in this class.
    """
    def __init__(self, driver, bot):
        """
        Constructor for ScrapingLogic. Will be called from the bot.

        :param driver: the selenium driver
        :param bot: the bot that is using the driver
        """
        self._driver = driver
        self._bot = bot

    def __getitem__(self, item):
        result = self._driver[item]
        if isinstance(result, type(self._driver)):
            result = ScrapingLogic(result)
        return result

    def __getattr__(self, item):
        result = getattr(self._driver, item)
        if callable(result):
            result = tocontainer(result, self._bot)
        return result

    def __repr__(self):
        return repr(self._driver)

    def sleep(self, seconds):
        """
        Sleep for a given amount of seconds.
        :param seconds:  the amount of seconds to sleep
        :return:  None
        """
        time.sleep(seconds)

    def replace_input_text(self, selector, keys):
        """
        If you have an input field and you want to replace the text in it, you can use this function.. It will first
        scroll to the element, then select all text in the input field and then send the new keys to the input field.

        :param selector: CSS or XPATH selector
        :param keys:  the new text that should be in the input field
        :return: True if the operation was successful, False otherwise
        """
        e = self.get_best_element(selector)
        if e is not None:
            try:
                self.scroll_to_element_by_js(e)
                self.sleep(1)
                e.send_keys(Keys.LEFT_CONTROL + "a")
                e.send_keys(keys)
            except Exception as e:
                pass
            return True
        return False

    def clear_input_text(self, selector):
        """
        If you have an input field and you want to clear the text in it, you can use this function. It will first
        scroll to the element, then select all text in the input field and then send the DELETE key to the input field.

        :param selector:  CSS or XPATH selector
        :return:  True if the operation was successful, False otherwise
        """
        e = self.get_best_element(selector)
        if e is not None:
            try:
                self.scroll_to_element_by_js(e)
                self.sleep(1)
                e.clear()
            except Exception as e:
                pass

            try:
                self.sleep(1)
                self.scroll_to_element_by_js(e)
                self.sleep(1)
                e.focus()
                e.send_keys(Keys.Control + "a")
                e.send_keys(Keys.DELETE)
                e.blur()
            except:
                pass
            return True
        return False

    def send_keys_to_element(self, selector, keys):
        """
        Send keys to an element. This function will send the keys to a CSS or XPATH element.

        :param selector:  CSS or XPATH selector
        :param keys: the keys that should be sent to the element
        :return: True if the operation was successful, False otherwise
        """
        e = self.get_best_element(selector)

        if e is not None:
            e.send_keys(keys)
            return True
        return False


    def is_visible(self, selector):
        e = self.get_best_element(selector)
        if e is None:
            return False
        displayed = e.is_displayed()
        return displayed

    def set_data(self, key, value, send_to_backend=False):
        self._bot.set_data(key, value, send_to_backend=send_to_backend)

    def take_screenshot(self, step):
        try:
            self._bot._take_screenshot(step)
        except Exception as e:
            pass

    def append_data(self, key, value, send_to_backend=False):
        self._bot.append_data(key, value, send_to_backend=send_to_backend)

    def send_data_to_backend(self, key=None, data=None):
        return self._bot.send_data_to_backend(key, data)

    def has_data(self, key):
        return self._bot.has_data(key)

    def get_data(self, key):
        return self._bot.get_data(key)

    def notify(self, message):
        self._bot.backend_notify(message)

    def convert_table_to_df(self, t):
        df = pd.read_html(t.get_attribute("outerHTML"))
        if len(df) > 0:
            df = df[0]
        return df

    def get_number_of_content(self, selector):

        try:
            e = self.get_best_element(selector)
            if e is not None:
                return int(e.get_attribute("innerText"))
        except:
            pass
        return None

    def convert_tables_to_df(self, tables):
        dfs = []
        for table in tables:
            df = self.convert_table_to_df(table)
            dfs.append(df)
        if len(dfs) <= 0:
            return None
        if len(dfs) == 1:
            return dfs[0]
        return dfs

    def get_all_elements(self, selector):
        by = By.CSS_SELECTOR
        if selector.startswith("//"):
            by = By.XPATH
        try:
            e = self._driver.find_elements(by, selector)
            return e
        except:
            pass
        return []

    def wait_for_reload(self, timeout=40, min_wait=0.1):
        was_incomplete = False
        # set a windows javascript variable to true if the page is still loading
        self._driver.execute_script("window._incomplete_data = true")
        for i in range(int(timeout/min_wait)):
            time.sleep(min_wait)
            # check if the variable is still true
            was_incomplete = self._driver.execute_script("return window._incomplete_data")
            # if the page is not loading anymore, return
            if not was_incomplete:
                # when document readystate
                if self._driver.execute_script("return document.readyState") == "complete":
                    return True
        return False

    def get_best_element(self, selector):
        by = By.CSS_SELECTOR
        if selector.startswith("//"):
            by = By.XPATH
        try:
            e = self._driver.find_element(by, selector)
            return e
        except:
            pass
        return None

    def element_count(self, selector):
        by = By.CSS_SELECTOR
        if selector.startswith("//"):
            by = By.XPATH
        try:
            e = self._driver.find_elements(by, selector)
            return len(e)
        except:
            return 0

    def element_text(self, selector):
        """
        Get the text of an element. This function will return the text of the element if it is available.

        :param selector: CSS or XPATH selector
        :return:  the text of the element
        """
        def get_best_match( elements):
            """
            Get the best match from a list of elements. This function will return the longest element from the list.

            :param elements:  a list of elements
            :return:  the longest element
            """
            longest = ""
            for e in elements:
                if len(e) > len(longest):
                    longest = e
            return longest
        by = By.CSS_SELECTOR
        if selector.startswith("//"):
            by = By.XPATH

        e = self._driver.find_element(by, selector)
        if e is not None:
            elements = []

            if e.text is not None:
                elements.append(e.text)
            if e.get_attribute("data-value") is not None:
                elements.append(e.get_attribute("data-value"))
            if e.get_attribute("value") is not None:
                elements.append(e.get_attribute("value"))
            if e.get_attribute("innerHTML") is not None:
                elements.append(e.get_attribute("innerHTML"))
            if e.get_attribute("innerText") is not None:
                elements.append(e.get_attribute("innerText"))
            if e.get_attribute("textContent") is not None:
                elements.append(e.get_attribute("textContent"))
            if len(elements) > 0:
                return get_best_match(elements)
        return None

    def wait_until_present(self, selector, timeout=20):
        """
        Wait until an element is present. This function will wait until an element is present in the DOM.

        :param selector:  CSS or XPATH selector
        :param timeout: the timeout in seconds
        :return: True if the element is present, False otherwise
        """
        if isinstance(selector, str):
            e = self.get_best_element(selector)
        else:
            e = selector
        if e is not None:
            try:
                wait = WebDriverWait(self._driver, timeout)
                wait.until(EC.presence_of_element_located(e))
                return True
            except:
                pass
        return False

    def wait_until_clickable(self, selector, timeout=10):
        if isinstance(selector, str):
            e = self.get_best_element(selector)
        else:
            e = selector
        if e is not None:
            try:
                wait = WebDriverWait(self._driver, timeout)
                wait.until(EC.element_to_be_clickable(e))
                return True
            except:
                pass

        return False

    def element_exists(self, selector):
        try:
            e = self.get_all_elements(selector)
            if e is not None:
                if len(e) > 0:
                    return True
        except:
            return False
        return False

    def set_attribute(self, selector, attribute, value):
        if isinstance(selector, str):
            e = self.get_best_element(selector)
        else:
            e = selector
        if e is not None:
            e.set_attribute(attribute, value)
            return True
        return False

    def scroll_to_element(self, selector):
        if isinstance(selector, str):
            e = self.get_best_element(selector)
        else:
            e = selector
        if e is not None:
            self._driver.execute_script("arguments[0].scrollIntoView();", e)
            return True
        return False

    def click_on_element_by_xpath_with_jquery(self, xpath):
        self._driver.execute_script(
            "$(document.evaluate(arguments[0], document, null, XPathResult.FIRST_ORDERED_NODE_TYPE, null).singleNodeValue)[0].click();",
            xpath)

    def inner_text_contains(self, selector, text):
        b = self.get_best_element(selector)
        if b is not None:
            innerTExt = b.get_attribute("innerText")
            if innerTExt is not None:
                return innerTExt.find(text) >= 0
        return False

    def click_on_best_element(self, selector):
        e = self.get_best_element(selector)
        if e is not None:
            e.click()
            return True
        return False

    def click_by_jquery_on_node(self, parent_button):
        if isinstance(parent_button, str):
            parent_button = self.get_best_element(parent_button)
        self._driver.execute_script("$(arguments[0]).click();", parent_button)
        return True

    def scroll_to_element_by_js(self, object):
        if isinstance(object, str):
            object = self.get_best_element(object)
        self._driver.execute_script(
            "window.scrollTo(arguments[0].getBoundingClientRect().left, arguments[0].getBoundingClientRect().top - (window.innerHeight/2) + document.documentElement.scrollTop);",
            object)
        return True